@@ -1,3 +1,12 @@
+try:
+    # pre-load required packages if we are running in pyodide.
+    # if we don't do this, we get import errors even though the modules are importable from the IPython console.
+    import micropip
+    micropip.install(['sympy', 'typing_extensions', 'plotly', 'scipy', 'tqdm'])
+except ImportError:
+    # if micropip isn't available, we're not running in pyodide.
+    pass
+
 import sympy
 
 from mathpad.val import *
@@ -26,4 +35,4 @@ except ImportError:
 sympy.init_printing()  # type: ignore
 sympy.printing.printer.Printer.set_global_settings(min=-3, max=4)  # type: ignore
 
-__version__ = "1.0.0"
+__version__ = "1.1.0"
